@@ -11,7 +11,7 @@
 #                         [-make <make prog>]
 #                        (and others)
 
-# Copyright (C) 1992-2019 Free Software Foundation, Inc.
+# Copyright (C) 1992-2022 Free Software Foundation, Inc.
 # This file is part of GNU Make.
 #
 # GNU Make is free software; you can redistribute it and/or modify it under
@@ -25,24 +25,30 @@
 # details.
 #
 # You should have received a copy of the GNU General Public License along with
-# this program.  If not, see <http://www.gnu.org/licenses/>.
+# this program.  If not, see <https://www.gnu.org/licenses/>.
 
 # Add the working directory to @INC and load the test driver
 use FindBin;
 use lib "$FindBin::Bin";
 
+our $testsroot = $FindBin::Bin;
+
 require "test_driver.pl";
 
-use File::Spec::Functions qw(:DEFAULT splitdir splitpath catpath);
+use File::Spec;
 
 use Cwd;
 $cwdpath = cwd();
-($cwdvol, $cwddir, $_) = splitpath($cwdpath, 1);
+($cwdvol, $cwddir, $_) = File::Spec->splitpath($cwdpath, 1);
 
 # Some target systems might not have the POSIX module...
 $has_POSIX = eval { require "POSIX.pm" };
 
 %FEATURES = ();
+%DEFVARS = (
+    AR => undef,
+    CC => undef
+);
 
 $valgrind = 0;              # invoke make with valgrind
 $valgrind_args = '';
@@ -52,12 +58,17 @@ $pure_log = undef;
 
 # The location of the GNU make source directory
 $srcdir = undef;
+$fqsrcdir = undef;
 $srcvol = undef;
 
 # The location of the build directory
 $blddir = undef;
+$fqblddir = undef;
 $bldvol = undef;
 
+$make_path = undef;
+@make_command = ();
+
 $command_string = '';
 
 $all_tests = 0;
@@ -95,8 +106,9 @@ if ($^O eq 'VMS')
 $ERR_no_such_file = undef;
 $ERR_read_only_file = undef;
 $ERR_unreadable_file = undef;
-$ERR_noexe_file = undef;
+$ERR_nonexe_file = undef;
 $ERR_exe_dir = undef;
+$ERR_command_not_found = undef;
 
 {
   use locale;
@@ -107,6 +119,9 @@ $ERR_exe_dir = undef;
       # Windows has POSIX locale, but only LC_ALL not LC_MESSAGES
       $loc = POSIX::setlocale(&POSIX::LC_ALL);
       POSIX::setlocale(&POSIX::LC_ALL, 'C');
+
+      # See set_defaults() as this doesn't work right on Windows :(
+      $! = &POSIX::ERANGE;
   }
 
   if (open(my $F, '<', 'file.none')) {
@@ -126,14 +141,14 @@ $ERR_exe_dir = undef;
       $ERR_read_only_file = "$!";
   }
 
-  $_ = `./file.out 2>/dev/null`;
+  $_ = `./file.out 2>&1`;
   if ($? == 0) {
       print "Executed non-executable file!  Skipping related tests.\n";
   } else {
       $ERR_nonexe_file = "$!";
   }
 
-  $_ = `./. 2>/dev/null`;
+  $_ = `./. 2>&1`;
   if ($? == 0) {
       print "Executed directory!  Skipping related tests.\n";
   } else {
@@ -150,6 +165,15 @@ $ERR_exe_dir = undef;
 
   unlink('file.out') or die "Failed to delete file.out: $!\n";
 
+  $_ = `/bin/sh -c 'bad-command 2>&1'`;
+  if ($? == 0) {
+      print "Invoked invalid file!  Skipping related tests.\n";
+  } else {
+      chomp($_);
+      s/bad-command/#CMDNAME#/g;
+      $ERR_command_not_found = $_;
+  }
+
   $loc and POSIX::setlocale(&POSIX::LC_ALL, $loc);
 }
 
@@ -161,18 +185,18 @@ sub valid_option
 
    if ($option =~ /^-make([-_]?path)?$/i) {
        $make_path = shift @argv;
-       if (!-f $make_path) {
+       if (! -f $make_path) {
            print "$option $make_path: Not found.\n";
-           exit 0;
+           exit 1;
        }
        return 1;
    }
 
    if ($option =~ /^-srcdir$/i) {
        $srcdir = shift @argv;
-       if (! -f catfile($srcdir, 'src', 'gnumake.h')) {
+       if (! -f File::Spec->catfile($srcdir, 'src', 'gnumake.h')) {
            print "$option $srcdir: Not a valid GNU make source directory.\n";
-           exit 0;
+           exit 1;
        }
        return 1;
    }
@@ -213,7 +237,11 @@ sub valid_option
 #  [2] (string):  Answer we should get back.
 #  [3] (integer): Exit code we expect.  A missing code means 0 (success)
 
+$makefile = undef;
 $old_makefile = undef;
+$mkpath = undef;
+$make_name = undef;
+$helptool = undef;
 
 sub subst_make_string
 {
@@ -223,6 +251,9 @@ sub subst_make_string
     s/#MAKE#/$make_name/g;
     s/#PERL#/$perl_name/g;
     s/#PWD#/$cwdpath/g;
+    s/#WORK#/$workdir/g;
+    # If we're using a shell
+    s/#HELPER#/$perl_name $helptool/g;
     return $_;
 }
 
@@ -239,7 +270,7 @@ sub run_make_test
     defined $old_makefile
       or die "run_make_test(undef) invoked before run_make_test('...')\n";
     $makefile = $old_makefile;
-  } else {
+  } elsif ($makestring) {
     if (! defined($makefile)) {
       $makefile = &get_tmpfile();
     }
@@ -283,7 +314,7 @@ sub add_options {
 }
 
 sub create_command {
-  return !$_[0] || ref($_[0]) ? [$make_path] : $make_path;
+  return !$_[0] || ref($_[0]) ? [@make_command] : join(' ', @make_command);
 }
 
 # The old-fashioned way...
@@ -293,7 +324,7 @@ sub create_command {
 # Using a ref should be preferred as it's more portable but all the older
 # invocations use strings.
 sub run_make_with_options {
-  my ($filename,$options,$logname,$expected_code,$timeout,@call) = @_;
+  my ($filename, $options, $logname, $expected_code, $timeout, @call) = @_;
   @call = caller unless @call;
   my $code;
   my $command = create_command($options);
@@ -347,13 +378,8 @@ sub run_make_with_options {
     $command = add_options($command, $options);
   }
 
-  my $cmdstr = ref($command) ? "'".join("' '", @$command)."'" : $command;
-
-  if (@call) {
-    $command_string = "#$call[1]:$call[2]\n$cmdstr\n";
-  } else {
-    $command_string = $cmdstr;
-  }
+  my $cmdstr = cmd2str($command);
+  $command_string = "# $call[1]:$call[2]:\n$cmdstr\n";
 
   if ($valgrind) {
     print VALGRIND "\n\nExecuting: $cmdstr\n";
@@ -384,7 +410,7 @@ sub run_make_with_options {
       # If we have a purify log, save it
       $tn = $pure_testname . ($num_of_logfiles ? ".$num_of_logfiles" : "");
       print("Renaming purify log file to $tn\n") if $debug;
-      rename($pure_log, "$tn") or die "Can't rename $log to $tn: $!\n";
+      rename($pure_log, "$tn") or die "Can't rename $pure_log to $tn: $!\n";
       ++$purify_errors;
     } else {
       unlink($pure_log);
@@ -392,10 +418,9 @@ sub run_make_with_options {
   }
 
   if ($code != $expected_code) {
-    print "Error running $make_path (expected $expected_code; got $code): $cmdstr\n";
+    print "Error running @make_command (expected $expected_code; got $code)\n$call[1]:$call[2]: $cmdstr\n";
     $test_passed = 0;
-    $runf = &get_runfile;
-    &create_file (&get_runfile, $command_string);
+    &create_file(get_runfile(), $command_string);
     # If it's a SIGINT, stop here
     if ($code & 127) {
       print STDERR "\nCaught signal ".($code & 127)."!\n";
@@ -405,7 +430,7 @@ sub run_make_with_options {
   }
 
   if ($profile & $vos) {
-    system "add_profile $make_path";
+    system "add_profile @make_command";
   }
 
   return 1;
@@ -437,7 +462,7 @@ sub print_help
 sub set_defaults
 {
   # $profile = 1;
-  $testee = "GNU make";
+  $testee = "GNU Make";
   $make_path = "make";
   $tmpfilesuffix = "mk";
   if ($port_type eq 'UNIX') {
@@ -464,32 +489,37 @@ sub set_defaults
 sub find_prog
 {
   my $prog = $_[0];
-  my ($v, $d, $f) = splitpath($prog);
+  my ($v, $d, $f) = File::Spec->splitpath($prog);
 
   # If there's no directory then we need to search the PATH
   if (! $d) {
-    foreach my $e (path()) {
-      $prog = catfile($e, $f);
-      -x $prog or continue;
-      ($v, $d, $f) = splitpath($prog);
-      last;
+    foreach my $e (File::Spec->path()) {
+      $prog = File::Spec->catfile($e, $f);
+      if (-x $prog) {
+        ($v, $d, $f) = File::Spec->splitpath($prog);
+        last;
+      }
     }
   }
 
   return ($v, $d, $f);
 }
 
+sub get_config
+{
+  return exists($CONFIG_FLAGS{$_[0]}) ? $CONFIG_FLAGS{$_[0]} : '';
+}
+
 sub set_more_defaults
 {
-  local($string);
-  local($index);
+  my $string;
 
   # Now that we have located make_path, locate the srcdir and blddir
   my ($mpv, $mpd, $mpf) = find_prog($make_path);
 
   # We have a make program so try to compute the blddir.
   if ($mpd) {
-    my $f = catpath($mpv, catdir($mpd, 'tests'), 'config-flags.pm');
+    my $f = File::Spec->catpath($mpv, File::Spec->catdir($mpd, 'tests'), 'config-flags.pm');
     if (-f $f) {
       $bldvol = $mpv;
       $blddir = $mpd;
@@ -499,7 +529,7 @@ sub set_more_defaults
   # If srcdir wasn't provided on the command line, try to find it.
   if (! $srcdir && $blddir) {
     # See if the blddir is the srcdir
-    my $f = catpath($bldvol, catdir($blddir, 'src'), 'gnumake.h');
+    my $f = File::Spec->catpath($bldvol, File::Spec->catdir($blddir, 'src'), 'gnumake.h');
     if (-f $f) {
       $srcdir = $blddir;
       $srcvol = $bldvol;
@@ -508,9 +538,9 @@ sub set_more_defaults
 
   if (! $srcdir) {
     # Not found, see if our parent is the source dir
-    my $f = catpath($cwdvol, catdir(updir(), 'src'), 'gnumake.h');
+    my $f = File::Spec->catpath($cwdvol, File::Spec->catdir(File::Spec->updir(), 'src'), 'gnumake.h');
     if (-f $f) {
-      $srcdir = updir();
+      $srcdir = File::Spec->updir();
       $srcvol = $cwdvol;
     }
   }
@@ -525,20 +555,15 @@ sub set_more_defaults
   if (!$blddir) {
     warn "Cannot locate config-flags.pm (no blddir)\n";
   } else {
-    my $f = catpath($bldvol, catdir($blddir, 'tests'), 'config-flags.pm');
+    my $f = File::Spec->catpath($bldvol, File::Spec->catdir($blddir, 'tests'), 'config-flags.pm');
     if (! -f $f) {
       warn "Cannot locate $f\n";
     } else {
-      unshift(@INC, catpath($bldvol, catdir($blddir, 'tests'), ''));
+      unshift(@INC, File::Spec->catpath($bldvol, File::Spec->catdir($blddir, 'tests'), ''));
       require "config-flags.pm";
     }
   }
 
-  # On DOS/Windows system the filesystem apparently can't track
-  # timestamps with second granularity (!!).  Change the sleep time
-  # needed to force a file to be considered "old".
-  $wtime = $port_type eq 'UNIX' ? 1 : $port_type eq 'OS/2' ? 2 : 4;
-
   # Find the full pathname of Make.  For DOS systems this is more
   # complicated, so we ask make itself.
   if ($osname eq 'VMS') {
@@ -549,17 +574,17 @@ sub set_more_defaults
     create_file('make.mk', 'all:;$(info $(MAKE))');
     my $mk = `$make_path -sf make.mk`;
     unlink('make.mk');
-    chop $mk;
+    $mk =~ s/\r?\n$//;
     $mk or die "FATAL ERROR: Cannot determine the value of \$(MAKE)\n";
     $make_path = $mk;
   }
-  ($mpv, $mpd, $mpf) = splitpath($make_path);
+  ($mpv, $mpd, $mpf) = File::Spec->splitpath($make_path);
 
   # Ask make what shell to use
   create_file('shell.mk', 'all:;$(info $(SHELL))');
   $sh_name = `$make_path -sf shell.mk`;
   unlink('shell.mk');
-  chop $sh_name;
+  $sh_name =~ s/\r?\n$//;
   if (! $sh_name) {
       print "Cannot determine shell\n";
       $is_posix_sh = 0;
@@ -587,21 +612,33 @@ sub set_more_defaults
   # start with a slash, but contains one).  Thanks for the
   # clue, Roland.
 
-  if ($mpd && !file_name_is_absolute($make_path) && $cwdvol == $mpv) {
-     $mkpath = catpath($cwdvol, catdir($cwd, $mpd), $mpf);
+  if ($mpd && !File::Spec->file_name_is_absolute($make_path) && $cwdvol == $mpv) {
+     $mkpath = File::Spec->catpath($cwdvol, File::Spec->catdir($cwd, $mpd), $mpf);
   } else {
      $mkpath = $make_path;
   }
 
   # Not with the make program, so see if we can get it out of the makefile
-  if (! $srcdir && open(MF, '<', catfile(updir(), 'Makefile'))) {
+  if (! $srcdir && open(MF, '<', File::Spec->catfile(File::Spec->updir(), 'Makefile'))) {
     local $/ = undef;
     $_ = <MF>;
     close(MF);
     /^abs_srcdir\s*=\s*(.*?)\s*$/m;
-    -f catfile($1, 'src', 'gnumake.h') and $srcdir = $1;
+    -f File::Spec->catfile($1, 'src', 'gnumake.h') and $srcdir = $1;
   }
 
+  # At this point we should have srcdir and blddir: get fq versions
+  $fqsrcdir = File::Spec->rel2abs($srcdir);
+  $fqblddir = File::Spec->rel2abs($blddir);
+
+  # Find the helper tool
+  $helptool = File::Spec->catfile($fqsrcdir, 'tests', 'thelp.pl');
+
+  # It's difficult to quote this properly in all the places it's used so
+  # ensure it doesn't need to be quoted.
+  $helptool =~ s,\\,/,g if $port_type eq 'W32';
+  $helptool =~ s, ,\\ ,g;
+
   # Get Purify log info--if any.
 
   if (exists $ENV{PURIFYOPTIONS}
@@ -625,16 +662,30 @@ sub set_more_defaults
   %FEATURES = map { $_ => 1 } split /\s+/, `$make_path -sf features.mk`;
   unlink('features.mk');
 
+  # Find the default values for different built-in variables
+  my $s = "all:;\n";
+  foreach (keys %DEFVARS) {
+      $s .= "\$(info $_=\$($_))\n";
+  }
+  create_file('defvars.mk', $s);
+  foreach (split "\n", `$make_path -sf defvars.mk`) {
+      my @e = split /=/, $_, 2;
+      $DEFVARS{$e[0]} = $e[1];
+  }
+  unlink('defvars.mk');
+
   # Set up for valgrind, if requested.
 
-  $make_command = $make_path;
+  @make_command = ($make_path);
 
   if ($valgrind) {
     my $args = $valgrind_args;
     open(VALGRIND, "> valgrind.out") or die "Cannot open valgrind.out: $!\n";
     #  -q --leak-check=yes
     exists $ENV{VALGRIND_ARGS} and $args = $ENV{VALGRIND_ARGS};
-    $make_path = "valgrind --log-fd=".fileno(VALGRIND)." $args $make_path";
+    @make_command = ('valgrind', '--log-fd='.fileno(VALGRIND));
+    push(@make_command, split(' ', $args));
+    push(@make_command, $make_path);
     # F_SETFD is 2
     fcntl(VALGRIND, 2, 0) or die "fcntl(setfd) failed: $!\n";
     system("echo Starting on `date` 1>&".fileno(VALGRIND));
@@ -642,13 +693,13 @@ sub set_more_defaults
   }
 
   if ($debug) {
-    print "Port type:  $port_type\n";
-    print "Make path:  $make_path\n";
-    print "Shell path: $sh_name".($is_posix_sh ? ' (POSIX)' : '')."\n";
-    print "#PWD#:      $cwdpath\n";
-    print "#PERL#:     $perl_name\n";
-    print "#MAKEPATH#: $mkpath\n";
-    print "#MAKE#:     $make_name\n";
+    print "Port type:    $port_type\n";
+    print "Make command: @make_command\n";
+    print "Shell path:   $sh_name".($is_posix_sh ? ' (POSIX)' : '')."\n";
+    print "#PWD#:        $cwdpath\n";
+    print "#PERL#:       $perl_name\n";
+    print "#MAKEPATH#:   $mkpath\n";
+    print "#MAKE#:       $make_name\n";
   }
 }
 
